@@ -156,9 +156,20 @@ static const struct gprs_rlc_cfg rlc_cfg_default = {
 
 static int gsm_bts_talloc_destructor(struct gsm_bts *bts)
 {
+	/* Entries in bts->loc_list are freed by talloc recursively, no need to free them here. */
+
+	llist_del(&bts->list);
+	hash_del(&bts->node_by_nr);
+	hash_del(&bts->node_by_lac);
+	hash_del(&bts->node_by_lac_ci);
+	hash_del(&bts->node_by_ci);
+
 	paging_destructor(bts);
 	bts_setup_ramp_remove(bts);
 
+	acc_ramp_deinit(&bts->acc_ramp);
+	acc_mgr_deinit(&bts->acc_mgr);
+
 	osmo_timer_del(&bts->cbch_timer);
 
 	bts->site_mgr->bts[0] = NULL;
@@ -182,7 +193,7 @@ static int gsm_bts_talloc_destructor(struct gsm_bts *bts)
  * This part is shared among the thin programs in osmo-bsc/src/utils/.
  * osmo-bsc requires further initialization that pulls in more dependencies (see
  * bsc_bts_alloc_register()). */
-struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm, uint8_t bts_num)
+struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm, gsm_bts_nr_t bts_num)
 {
 	struct gsm_bts *bts = talloc_zero(bts_sm, struct gsm_bts);
 
@@ -192,6 +203,17 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	talloc_set_destructor(bts, gsm_bts_talloc_destructor);
 
 	bts->nr = bts_num;
+
+	llist_add_tail(&bts->list, &net->bts_list);
+	hash_add(net->bts_by_nr, &bts->node_by_nr, bts->nr);
+	net->num_bts++;
+
+	/* Default bts->location_area_code == GSM_LAC_RESERVED_DETACHED, don't add to hashtable: */
+	INIT_HLIST_NODE(&bts->node_by_lac);
+	INIT_HLIST_NODE(&bts->node_by_lac_ci);
+	/* Default CI = 0: */
+	hash_add(net->bts_by_ci, &bts->node_by_ci, bts->cell_identity);
+
 	bts->num_trx = 0;
 	INIT_LLIST_HEAD(&bts->trx_list);
 	bts->network = net;
@@ -203,7 +225,7 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	bts->mo.fi = osmo_fsm_inst_alloc(&nm_bts_fsm, bts, bts,
 					      LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(bts->mo.fi, "bts%d", bts->nr);
-	gsm_mo_init(&bts->mo, bts, NM_OC_BTS, bts->nr, 0xff, 0xff);
+	gsm_mo_init(&bts->mo, bts, NM_OC_BTS, bts->bts_nr, 0xff, 0xff);
 
 	/* 3GPP TS 08.18, chapter 5.4.1: 0 is reserved for signalling */
 	bts->gprs.cell.bvci = 2;
@@ -215,7 +237,7 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 						   &bts->gprs.cell, LOGL_INFO, NULL);
 	osmo_fsm_inst_update_id_f(bts->gprs.cell.mo.fi, "gprs-cell%d", bts->nr);
 	gsm_mo_init(&bts->gprs.cell.mo, bts, NM_OC_GPRS_CELL,
-			bts->nr, 0xff, 0xff);
+		    bts->bts_nr, 0xff, 0xff);
 
 	/* init statistics */
 	bts->bts_ctrs = rate_ctr_group_alloc(bts, &bts_ctrg_desc, bts->nr);
@@ -343,6 +365,7 @@ struct gsm_bts *gsm_bts_alloc(struct gsm_network *net, struct gsm_bts_sm *bts_sm
 	INIT_LLIST_HEAD(&bts->neighbors);
 	INIT_LLIST_HEAD(&bts->oml_fail_rep);
 	INIT_LLIST_HEAD(&bts->chan_rqd_queue);
+	INIT_LLIST_HEAD(&bts->depends_on);
 
 	/* Don't pin the BTS to any MGW by default: */
 	bts->mgw_pool_target = -1;
@@ -502,9 +525,9 @@ __attribute__((weak)) int gsm_bts_check_cfg(struct gsm_bts *bts)
 		}
 		break;
 	case GSM_BAND_900:
-		if ((bts->c0->arfcn > 124 && bts->c0->arfcn < 955) ||
+		if ((bts->c0->arfcn > 124 && bts->c0->arfcn < 940) ||
 		    bts->c0->arfcn > 1023) {
-			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM900 channel (%u) must be between 0-124, 955-1023.\n",
+			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM900 channel (%u) must be between 0-124, 940-1023.\n",
 			     bts->nr, bts->c0->arfcn);
 			return -EINVAL;
 		}
@@ -735,6 +758,30 @@ int gsm_set_bts_type(struct gsm_bts *bts, enum gsm_bts_type type)
 	return 0;
 }
 
+void gsm_bts_set_lac(struct gsm_bts *bts, uint16_t lac)
+{
+	bts->location_area_code = lac;
+
+	hash_del(&bts->node_by_lac);
+	hash_del(&bts->node_by_lac_ci);
+	hash_add(bts->network->bts_by_lac, &bts->node_by_lac, bts->location_area_code);
+	hash_add(bts->network->bts_by_lac_ci, &bts->node_by_lac_ci,
+		 LAC_CI_HASHTABLE_KEY(bts->location_area_code, bts->cell_identity));
+}
+
+void gsm_bts_set_ci(struct gsm_bts *bts, uint16_t ci)
+{
+	bts->cell_identity = ci;
+
+	hash_del(&bts->node_by_ci);
+	hash_add(bts->network->bts_by_ci, &bts->node_by_ci, bts->cell_identity);
+	if (bts->location_area_code != GSM_LAC_RESERVED_DETACHED) {
+		hash_del(&bts->node_by_lac_ci);
+		hash_add(bts->network->bts_by_lac_ci, &bts->node_by_lac_ci,
+			 LAC_CI_HASHTABLE_KEY(bts->location_area_code, bts->cell_identity));
+	}
+}
+
 int bts_gprs_mode_is_compat(struct gsm_bts *bts, enum bts_gprs_mode mode)
 {
 	if (mode != BTS_GPRS_NONE &&
@@ -869,37 +916,43 @@ void gsm_bts_mo_reset(struct gsm_bts *bts)
 	}
 }
 
-/* Assume there are only 256 possible bts */
-osmo_static_assert(sizeof(((struct gsm_bts *) 0)->nr) == 1, _bts_nr_is_256);
-static void depends_calc_index_bit(int bts_nr, int *idx, int *bit)
+int bts_depend_mark(struct gsm_bts *bts, gsm_bts_nr_t dep)
 {
-	*idx = bts_nr / (8 * 4);
-	*bit = bts_nr % (8 * 4);
+	struct bts_depends_on_entry *entry;
+	entry = talloc_zero(bts, struct bts_depends_on_entry);
+	if (!entry) {
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Alloc of struct bts_depends_on_entry failed");
+		return -1;
+	}
+	entry->bts_nr = dep;
+	llist_add_tail(&entry->list, &bts->depends_on);
+	return 0;
 }
 
-void bts_depend_mark(struct gsm_bts *bts, int dep)
+static struct bts_depends_on_entry *bts_depend_find_entry(const struct gsm_bts *bts, gsm_bts_nr_t dep)
 {
-	int idx, bit;
-	depends_calc_index_bit(dep, &idx, &bit);
-
-	bts->depends_on[idx] |= 1U << bit;
+	struct bts_depends_on_entry *entry;
+	llist_for_each_entry(entry, &bts->depends_on, list) {
+		if (entry->bts_nr == dep)
+			return entry;
+	}
+	return NULL;
 }
 
-void bts_depend_clear(struct gsm_bts *bts, int dep)
+void bts_depend_clear(struct gsm_bts *bts, gsm_bts_nr_t dep)
 {
-	int idx, bit;
-	depends_calc_index_bit(dep, &idx, &bit);
-
-	bts->depends_on[idx] &= ~(1U << bit);
+	struct bts_depends_on_entry *entry;
+	entry = bts_depend_find_entry(bts, dep);
+	if (!entry)
+		return;
+	llist_del(&entry->list);
+	talloc_free(entry);
 }
 
-int bts_depend_is_depedency(struct gsm_bts *base, struct gsm_bts *other)
+bool bts_depend_is_depedency(struct gsm_bts *base, struct gsm_bts *other)
 {
-	int idx, bit;
-	depends_calc_index_bit(other->nr, &idx, &bit);
-
-	/* Check if there is a depends bit */
-	return (base->depends_on[idx] & (1U << bit)) > 0;
+	struct bts_depends_on_entry *entry = bts_depend_find_entry(base, other->nr);
+	return !!entry;
 }
 
 static bool bts_is_online(const struct gsm_bts *bts)
@@ -914,7 +967,7 @@ static bool bts_is_online(const struct gsm_bts *bts)
 	return bts->mo.nm_state.operational == NM_OPSTATE_ENABLED;
 }
 
-int bts_depend_check(struct gsm_bts *bts)
+bool bts_depend_check(struct gsm_bts *bts)
 {
 	struct gsm_bts *other_bts;
 
@@ -923,9 +976,9 @@ int bts_depend_check(struct gsm_bts *bts)
 			continue;
 		if (bts_is_online(other_bts))
 			continue;
-		return 0;
+		return false;
 	}
-	return 1;
+	return true;
 }
 
 /* get the radio link timeout (based on SACCH decode errors, according
